from typing import NamedTuple
from string import ascii_letters


class UnknownLocaleError(Exception):
    pass


class Locale(NamedTuple):
    language: str
    territory: str = ''

    def __str__(self):
        if self.territory:
            return f'{self.language}_{self.territory}'
        return self.language

    @classmethod
    def parse(cls, identifier, sep):
        if not isinstance(identifier, str):
            raise TypeError(f"Unexpected value for identifier: '{identifier}'")
        locale = cls(*identifier.split(sep, 1))
        if not all(x in ascii_letters for x in locale.language):
            raise ValueError(f"expected only letters, got '{locale.language}'")
        if len(locale.language) != 2:
            raise UnknownLocaleError(f"unknown locale '{locale.language}'")
        return locale



from distutils.errors import DistutilsOptionError
from os import path
from pkg_resources import EntryPoint
from babel.messages import frontend as babel


DEFAULT_MAPPING_FILE = path.normpath(path.join(
    path.abspath(path.dirname(__file__)), '../themes/babel.cfg'
))


class ThemeMixin:
    def get_theme_dir(self):
        ''' Validate theme option and return path to theme's root obtained from entry point. '''
        entry_points = EntryPoint.parse_map(self.distribution.entry_points, self.distribution)
        if 'mkdocs.themes' not in entry_points:
            raise DistutilsOptionError("no mkdocs.themes are defined in entry_points")
        if self.theme is None and len(entry_points['mkdocs.themes']) == 1:
            # Default to the only theme defined in entry_points as none specified.
            self.theme = tuple(entry_points['mkdocs.themes'].keys())[0]
        if self.theme not in entry_points['mkdocs.themes']:
            raise DistutilsOptionError("you must specify a valid theme name to work on")
        theme = entry_points['mkdocs.themes'][self.theme]
        return path.dirname(theme.resolve().__file__)


class compile_catalog(babel.compile_catalog, ThemeMixin):
    user_options = babel.compile_catalog.user_options + [
        ("theme=", "t", "theme name to work on"),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.theme = None

    def finalize_options(self):
        if not self.directory:
            theme_dir = self.get_theme_dir()
            self.directory = f"{theme_dir}/locales"
        super().finalize_options()


class extract_messages(babel.extract_messages, ThemeMixin):
    user_options = babel.extract_messages.user_options + [
        ("domain=", "d", "domains of the POT output file"),
        ("theme=", "t", "theme name to work on"),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.domain = "messages"
        self.theme = None

    def finalize_options(self):
        if not self.version:
            version = self.distribution.get_version()
            self.version = ".".join([i for i in version.split(".") if "dev" not in i])
        if not self.mapping_file:
            self.mapping_file = DEFAULT_MAPPING_FILE
        if not self.input_paths or not self.output_file:
            theme_dir = self.get_theme_dir()
            if not self.input_paths:
                self.input_paths = theme_dir
            if not self.output_file:
                self.output_file = f"{theme_dir}/{self.domain}.pot"
        super().finalize_options()


class init_catalog(babel.init_catalog, ThemeMixin):
    user_options = babel.init_catalog.user_options + [
        ("theme=", "t", "theme name to work on"),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.theme = None

    def finalize_options(self):
        if not self.input_file or not self.output_dir:
            theme_dir = self.get_theme_dir()
            if not self.input_file:
                self.input_file = f"{theme_dir}/{self.domain}.pot"
            if not self.output_dir:
                self.output_dir = f"{theme_dir}/locales"
        super().finalize_options()


class update_catalog(babel.update_catalog, ThemeMixin):
    user_options = babel.update_catalog.user_options + [
        ("theme=", "t", "theme name to work on"),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.theme = None

    def finalize_options(self):
        if not self.input_file or not self.output_dir:
            theme_dir = self.get_theme_dir()
            if not self.input_file:
                self.input_file = f"{theme_dir}/{self.domain}.pot"
            if not self.output_dir:
                self.output_dir = f"{theme_dir}/locales"
        super().finalize_options()



#!/usr/bin/env python


import unittest
from unittest import mock
import os

from mkdocs import plugins
from mkdocs import config
from mkdocs.commands import build
from mkdocs.exceptions import BuildError, PluginError, Abort
from mkdocs.tests.base import load_config


class DummyPlugin(plugins.BasePlugin):
    config_scheme = (
        ('foo', config.config_options.Type(str, default='default foo')),
        ('bar', config.config_options.Type(int, default=0)),
        ('dir', config.config_options.Dir(exists=False)),
    )

    def on_pre_page(self, content, **kwargs):
        """ modify page content by prepending `foo` config value. """
        return '{} {}'.format(self.config['foo'], content)

    def on_nav(self, item, **kwargs):
        """ do nothing (return None) to not modify item. """
        return None

    def on_page_read_source(self, **kwargs):
        """ create new source by prepending `foo` config value to 'source'. """
        return '{} {}'.format(self.config['foo'], 'source')

    def on_pre_build(self, **kwargs):
        """ do nothing (return None). """
        return None


class TestPluginClass(unittest.TestCase):

    def test_valid_plugin_options(self):
        test_dir = 'test'

        options = {
            'foo': 'some value',
            'dir': test_dir,
        }

        cfg_fname = os.path.join('tmp', 'test', 'fname.yml')
        cfg_fname = os.path.abspath(cfg_fname)

        cfg_dirname = os.path.dirname(cfg_fname)
        expected = os.path.join(cfg_dirname, test_dir)

        expected = {
            'foo': 'some value',
            'bar': 0,
            'dir': expected,
        }

        plugin = DummyPlugin()
        errors, warnings = plugin.load_config(options, config_file_path=cfg_fname)
        self.assertEqual(plugin.config, expected)
        self.assertEqual(errors, [])
        self.assertEqual(warnings, [])

    def test_invalid_plugin_options(self):

        plugin = DummyPlugin()
        errors, warnings = plugin.load_config({'foo': 42})
        self.assertEqual(len(errors), 1)
        self.assertIn('foo', errors[0])
        self.assertEqual(warnings, [])

        errors, warnings = plugin.load_config({'bar': 'a string'})
        self.assertEqual(len(errors), 1)
        self.assertIn('bar', errors[0])
        self.assertEqual(warnings, [])

        errors, warnings = plugin.load_config({'invalid_key': 'value'})
        self.assertEqual(errors, [])
        self.assertEqual(len(warnings), 1)
        self.assertIn('invalid_key', warnings[0])


class TestPluginCollection(unittest.TestCase):

    def test_set_plugin_on_collection(self):
        collection = plugins.PluginCollection()
        plugin = DummyPlugin()
        collection['foo'] = plugin
        self.assertEqual([(k, v) for k, v in collection.items()], [('foo', plugin)])

    def test_set_multiple_plugins_on_collection(self):
        collection = plugins.PluginCollection()
        plugin1 = DummyPlugin()
        collection['foo'] = plugin1
        plugin2 = DummyPlugin()
        collection['bar'] = plugin2
        self.assertEqual([(k, v) for k, v in collection.items()], [('foo', plugin1), ('bar', plugin2)])

    def test_run_event_on_collection(self):
        collection = plugins.PluginCollection()
        plugin = DummyPlugin()
        plugin.load_config({'foo': 'new'})
        collection['foo'] = plugin
        self.assertEqual(collection.run_event('pre_page', 'page content'), 'new page content')

    def test_run_event_twice_on_collection(self):
        collection = plugins.PluginCollection()
        plugin1 = DummyPlugin()
        plugin1.load_config({'foo': 'new'})
        collection['foo'] = plugin1
        plugin2 = DummyPlugin()
        plugin2.load_config({'foo': 'second'})
        collection['bar'] = plugin2
        self.assertEqual(collection.run_event('pre_page', 'page content'),
                         'second new page content')

    def test_event_returns_None(self):
        collection = plugins.PluginCollection()
        plugin = DummyPlugin()
        plugin.load_config({'foo': 'new'})
        collection['foo'] = plugin
        self.assertEqual(collection.run_event('nav', 'nav item'), 'nav item')

    def test_event_empty_item(self):
        collection = plugins.PluginCollection()
        plugin = DummyPlugin()
        plugin.load_config({'foo': 'new'})
        collection['foo'] = plugin
        self.assertEqual(collection.run_event('page_read_source'), 'new source')

    def test_event_empty_item_returns_None(self):
        collection = plugins.PluginCollection()
        plugin = DummyPlugin()
        plugin.load_config({'foo': 'new'})
        collection['foo'] = plugin
        self.assertEqual(collection.run_event('pre_build'), None)

    def test_run_undefined_event_on_collection(self):
        collection = plugins.PluginCollection()
        self.assertEqual(collection.run_event('pre_page', 'page content'), 'page content')

    def test_run_unknown_event_on_collection(self):
        collection = plugins.PluginCollection()
        self.assertRaises(KeyError, collection.run_event, 'unknown', 'page content')

    def test_run_build_error_event(self):
        build_errors = []

        class PluginRaisingError(plugins.BasePlugin):
            def __init__(self, error_on):
                self.error_on = error_on

            def on_pre_page(self, page, **kwargs):
                if self.error_on == 'pre_page':
                    raise BuildError('pre page error')
                return page

            def on_page_markdown(self, markdown, **kwargs):
                if self.error_on == 'page_markdown':
                    raise BuildError('page markdown error')
                return markdown

            def on_page_content(self, html, **kwargs):
                if self.error_on == 'page_content':
                    raise PluginError('page content error')
                return html

            def on_post_page(self, html, **kwargs):
                if self.error_on == 'post_page':
                    raise ValueError('post page error')

            def on_build_error(self, error, **kwargs):
                build_errors.append(error)

        cfg = load_config()
        cfg['plugins']['errorplugin'] = PluginRaisingError(error_on='pre_page')
        self.assertRaises(Abort, build.build, cfg)

        cfg = load_config()
        cfg['plugins']['errorplugin'] = PluginRaisingError(error_on='page_markdown')
        self.assertRaises(Abort, build.build, cfg)

        cfg = load_config()
        cfg['plugins']['errorplugin'] = PluginRaisingError(error_on='page_content')
        self.assertRaises(Abort, build.build, cfg)

        cfg = load_config()
        cfg['plugins']['errorplugin'] = PluginRaisingError(error_on='post_page')
        self.assertRaises(ValueError, build.build, cfg)

        cfg = load_config()
        cfg['plugins']['errorplugin'] = PluginRaisingError(error_on='')
        build.build(cfg)

        self.assertEqual(len(build_errors), 4)
        self.assertIs(build_errors[0].__class__, BuildError)
        self.assertEqual(str(build_errors[0]), 'pre page error')
        self.assertIs(build_errors[1].__class__, BuildError)
        self.assertEqual(str(build_errors[1]), 'page markdown error')
        self.assertIs(build_errors[2].__class__, PluginError)
        self.assertEqual(str(build_errors[2]), 'page content error')
        self.assertIs(build_errors[3].__class__, ValueError)
        self.assertEqual(str(build_errors[3]), 'post page error')


MockEntryPoint = mock.Mock()
MockEntryPoint.configure_mock(**{'name': 'sample', 'load.return_value': DummyPlugin})


@mock.patch('importlib_metadata.entry_points', return_value=[MockEntryPoint])
class TestPluginConfig(unittest.TestCase):

    def test_plugin_config_without_options(self, mock_class):

        cfg = {'plugins': ['sample']}
        option = config.config_options.Plugins()
        cfg['plugins'] = option.validate(cfg['plugins'])

        self.assertIsInstance(cfg['plugins'], plugins.PluginCollection)
        self.assertIn('sample', cfg['plugins'])
        self.assertIsInstance(cfg['plugins']['sample'], plugins.BasePlugin)
        expected = {
            'foo': 'default foo',
            'bar': 0,
            'dir': None,
        }
        self.assertEqual(cfg['plugins']['sample'].config, expected)

    def test_plugin_config_with_options(self, mock_class):

        cfg = {
            'plugins': [{
                'sample': {
                    'foo': 'foo value',
                    'bar': 42
                }
            }]
        }
        option = config.config_options.Plugins()
        cfg['plugins'] = option.validate(cfg['plugins'])

        self.assertIsInstance(cfg['plugins'], plugins.PluginCollection)
        self.assertIn('sample', cfg['plugins'])
        self.assertIsInstance(cfg['plugins']['sample'], plugins.BasePlugin)
        expected = {
            'foo': 'foo value',
            'bar': 42,
            'dir': None,
        }
        self.assertEqual(cfg['plugins']['sample'].config, expected)

    def test_plugin_config_as_dict(self, mock_class):

        cfg = {
            'plugins': {
                'sample': {
                    'foo': 'foo value',
                    'bar': 42
                }
            }
        }
        option = config.config_options.Plugins()
        cfg['plugins'] = option.validate(cfg['plugins'])

        self.assertIsInstance(cfg['plugins'], plugins.PluginCollection)
        self.assertIn('sample', cfg['plugins'])
        self.assertIsInstance(cfg['plugins']['sample'], plugins.BasePlugin)
        expected = {
            'foo': 'foo value',
            'bar': 42,
            'dir': None,
        }
        self.assertEqual(cfg['plugins']['sample'].config, expected)

    def test_plugin_config_empty_list_with_empty_default(self, mock_class):
        cfg = {'plugins': []}
        option = config.config_options.Plugins(default=[])
        cfg['plugins'] = option.validate(cfg['plugins'])

        self.assertIsInstance(cfg['plugins'], plugins.PluginCollection)
        self.assertEqual(len(cfg['plugins']), 0)

    def test_plugin_config_empty_list_with_default(self, mock_class):
        # Default is ignored
        cfg = {'plugins': []}
        option = config.config_options.Plugins(default=['sample'])
        cfg['plugins'] = option.validate(cfg['plugins'])

        self.assertIsInstance(cfg['plugins'], plugins.PluginCollection)
        self.assertEqual(len(cfg['plugins']), 0)

    def test_plugin_config_none_with_empty_default(self, mock_class):
        cfg = {'plugins': None}
        option = config.config_options.Plugins(default=[])
        cfg['plugins'] = option.validate(cfg['plugins'])

        self.assertIsInstance(cfg['plugins'], plugins.PluginCollection)
        self.assertEqual(len(cfg['plugins']), 0)

    def test_plugin_config_none_with_default(self, mock_class):
        # Default is used.
        cfg = {'plugins': None}
        option = config.config_options.Plugins(default=['sample'])
        cfg['plugins'] = option.validate(cfg['plugins'])

        self.assertIsInstance(cfg['plugins'], plugins.PluginCollection)
        self.assertIn('sample', cfg['plugins'])
        self.assertIsInstance(cfg['plugins']['sample'], plugins.BasePlugin)
        expected = {
            'foo': 'default foo',
            'bar': 0,
            'dir': None,
        }
        self.assertEqual(cfg['plugins']['sample'].config, expected)

    def test_plugin_config_uninstalled(self, mock_class):

        cfg = {'plugins': ['uninstalled']}
        option = config.config_options.Plugins()
        self.assertRaises(config.base.ValidationError, option.validate, cfg['plugins'])

    def test_plugin_config_not_list(self, mock_class):

        cfg = {'plugins': 'sample'}  # should be a list
        option = config.config_options.Plugins()
        self.assertRaises(config.base.ValidationError, option.validate, cfg['plugins'])

    def test_plugin_config_multivalue_dict(self, mock_class):

        cfg = {
            'plugins': [{
                'sample': {
                    'foo': 'foo value',
                    'bar': 42
                },
                'extra_key': 'baz'
            }]
        }
        option = config.config_options.Plugins()
        self.assertRaises(config.base.ValidationError, option.validate, cfg['plugins'])

    def test_plugin_config_not_string_or_dict(self, mock_class):

        cfg = {
            'plugins': [('not a string or dict',)]
        }
        option = config.config_options.Plugins()
        self.assertRaises(config.base.ValidationError, option.validate, cfg['plugins'])

    def test_plugin_config_options_not_dict(self, mock_class):

        cfg = {
            'plugins': [{
                'sample': 'not a dict'
            }]
        }
        option = config.config_options.Plugins()
        self.assertRaises(config.base.ValidationError, option.validate, cfg['plugins'])



import os
import logging

from jinja2.ext import Extension, InternationalizationExtension
from mkdocs.config.base import ValidationError

try:
    from babel.core import Locale, UnknownLocaleError
    from babel.support import Translations, NullTranslations
    has_babel = True
except ImportError:  # pragma: no cover
    from mkdocs.utils.babel_stub import Locale, UnknownLocaleError
    has_babel = False


log = logging.getLogger(__name__)
base_path = os.path.dirname(os.path.abspath(__file__))


class NoBabelExtension(InternationalizationExtension):  # pragma: no cover
    def __init__(self, environment):
        Extension.__init__(self, environment)
        environment.extend(
            install_null_translations=self._install_null,
            newstyle_gettext=False,
        )


def parse_locale(locale):
    try:
        return Locale.parse(locale, sep='_')
    except (ValueError, UnknownLocaleError, TypeError) as e:
        raise ValidationError(f'Invalid value for locale: {str(e)}')


def install_translations(env, locale, theme_dirs):
    if has_babel:
        env.add_extension('jinja2.ext.i18n')
        translations = _get_merged_translations(theme_dirs, 'locales', locale)
        if translations is not None:
            env.install_gettext_translations(translations)
        else:
            env.install_null_translations()
            if locale.language != 'en':
                log.warning(
                    f"No translations could be found for the locale '{locale}'. "
                    'Defaulting to English.'
                )
    else:  # pragma: no cover
        # no babel installed, add dummy support for trans/endtrans blocks
        env.add_extension(NoBabelExtension)
        env.install_null_translations()


def _get_merged_translations(theme_dirs, locales_dir, locale):
    merged_translations = None

    log.debug(f"Looking for translations for locale '{locale}'")
    if locale.territory:
        locale_str = f"{locale.language}_{locale.territory}"
    else:
        locale_str = locale.language
    for theme_dir in reversed(theme_dirs):
        dirname = os.path.join(theme_dir, locales_dir)
        translations = Translations.load(dirname, [locale_str])

        if type(translations) is NullTranslations:
            log.debug(f"No translations found here: '{dirname}'")
            continue

        log.debug(f"Translations found here: '{dirname}'")
        if merged_translations is None:
            merged_translations = translations
        else:
            merged_translations.merge(translations)

    return merged_translations



import os
import logging
from mkdocs import utils
from mkdocs.plugins import BasePlugin
from mkdocs.config import config_options
from mkdocs.contrib.search.search_index import SearchIndex


log = logging.getLogger(__name__)
base_path = os.path.dirname(os.path.abspath(__file__))


class LangOption(config_options.OptionallyRequired):
    """ Validate Language(s) provided in config are known languages. """

    def get_lunr_supported_lang(self, lang):
        for lang_part in lang.split("_"):
            lang_part = lang_part.lower()
            if os.path.isfile(os.path.join(base_path, 'lunr-language', f'lunr.{lang_part}.js')):
                return lang_part

    def run_validation(self, value):
        if isinstance(value, str):
            value = [value]
        elif not isinstance(value, (list, tuple)):
            raise config_options.ValidationError('Expected a list of language codes.')
        for lang in list(value):
            if lang != 'en':
                lang_detected = self.get_lunr_supported_lang(lang)
                if not lang_detected:
                    log.info(f"Option search.lang '{lang}' is not supported, falling back to 'en'")
                    value.remove(lang)
                    if 'en' not in value:
                        value.append('en')
                elif lang_detected != lang:
                    value.remove(lang)
                    value.append(lang_detected)
                    log.info(f"Option search.lang '{lang}' switched to '{lang_detected}'")
        return value


class SearchPlugin(BasePlugin):
    """ Add a search feature to MkDocs. """

    config_scheme = (
        ('lang', LangOption()),
        ('separator', config_options.Type(str, default=r'[\s\-]+')),
        ('min_search_length', config_options.Type(int, default=3)),
        ('prebuild_index', config_options.Choice((False, True, 'node', 'python'), default=False)),
        ('indexing', config_options.Choice(('full', 'sections', 'titles'), default='full'))
    )

    def on_config(self, config, **kwargs):
        "Add plugin templates and scripts to config."
        if 'include_search_page' in config['theme'] and config['theme']['include_search_page']:
            config['theme'].static_templates.add('search.html')
        if not ('search_index_only' in config['theme'] and config['theme']['search_index_only']):
            path = os.path.join(base_path, 'templates')
            config['theme'].dirs.append(path)
            if 'search/main.js' not in config['extra_javascript']:
                config['extra_javascript'].append('search/main.js')
        if self.config['lang'] is None:
            # lang setting undefined. Set default based on theme locale
            validate = self.config_scheme[0][1].run_validation
            self.config['lang'] = validate(config['theme']['locale'].language)
        # The `python` method of `prebuild_index` is pending deprecation as of version 1.2.
        # TODO: Raise a deprecation warning in a future release (1.3?).
        if self.config['prebuild_index'] == 'python':
            log.info(
                "The 'python' method of the search plugin's 'prebuild_index' config option "
                "is pending deprecation and will not be supported in a future release."
            )
        return config

    def on_pre_build(self, config, **kwargs):
        "Create search index instance for later use."
        self.search_index = SearchIndex(**self.config)

    def on_page_context(self, context, **kwargs):
        "Add page to search index."
        self.search_index.add_entry_from_context(context['page'])

    def on_post_build(self, config, **kwargs):
        "Build search index."
        output_base_path = os.path.join(config['site_dir'], 'search')
        search_index = self.search_index.generate_search_index()
        json_output_path = os.path.join(output_base_path, 'search_index.json')
        utils.write_file(search_index.encode('utf-8'), json_output_path)

        if not ('search_index_only' in config['theme'] and config['theme']['search_index_only']):
            # Include language support files in output. Copy them directly
            # so that only the needed files are included.
            files = []
            if len(self.config['lang']) > 1 or 'en' not in self.config['lang']:
                files.append('lunr.stemmer.support.js')
            if len(self.config['lang']) > 1:
                files.append('lunr.multi.js')
            if ('ja' in self.config['lang'] or 'jp' in self.config['lang']):
                files.append('tinyseg.js')
            for lang in self.config['lang']:
                if (lang != 'en'):
                    files.append(f'lunr.{lang}.js')

            for filename in files:
                from_path = os.path.join(base_path, 'lunr-language', filename)
                to_path = os.path.join(output_base_path, filename)
                utils.copy_file(from_path, to_path)



"""
Standalone file utils.

Nothing in this module should have an knowledge of config or the layout
and structure of the site and pages in the site.
"""


import logging
import os
import shutil
import re
import yaml
import fnmatch
import posixpath
import functools
import importlib_metadata
from collections import defaultdict
from datetime import datetime, timezone
from urllib.parse import urlsplit
from yaml_env_tag import construct_env_tag
from mergedeep import merge

from mkdocs import exceptions

log = logging.getLogger(__name__)

markdown_extensions = [
    '.markdown',
    '.mdown',
    '.mkdn',
    '.mkd',
    '.md'
]


def get_yaml_loader(loader=yaml.Loader):
    """ Wrap PyYaml's loader so we can extend it to suit our needs. """

    class Loader(loader):
        """
        Define a custom loader derived from the global loader to leave the
        global loader unaltered.
        """

    # Attach Environment Variable constructor.
    # See https://github.com/waylan/pyyaml-env-tag
    Loader.add_constructor('!ENV', construct_env_tag)

    return Loader


def yaml_load(source, loader=None):
    """ Return dict of source YAML file using loader, recursively deep merging inherited parent. """
    Loader = loader or get_yaml_loader()
    result = yaml.load(source, Loader=Loader)
    if result is not None and 'INHERIT' in result:
        relpath = result.pop('INHERIT')
        abspath = os.path.normpath(os.path.join(os.path.dirname(source.name), relpath))
        if not os.path.exists(abspath):
            raise exceptions.ConfigurationError(
                f"Inherited config file '{relpath}' does not exist at '{abspath}'.")
        log.debug(f"Loading inherited configuration file: {abspath}")
        with open(abspath, 'rb') as fd:
            parent = yaml_load(fd, Loader)
        result = merge(parent, result)
    return result


def modified_time(file_path):
    """
    Return the modified time of the supplied file. If the file does not exists zero is returned.
    see build_pages for use.
    """
    if os.path.exists(file_path):
        return os.path.getmtime(file_path)
    else:
        return 0.0


def get_build_timestamp():
    """
    Returns the number of seconds since the epoch.

    Support SOURCE_DATE_EPOCH environment variable for reproducible builds.
    See https://reproducible-builds.org/specs/source-date-epoch/
    """
    source_date_epoch = os.environ.get('SOURCE_DATE_EPOCH')
    if source_date_epoch is None:
        return int(datetime.now(timezone.utc).timestamp())

    return int(source_date_epoch)


def get_build_datetime():
    """
    Returns an aware datetime object.

    Support SOURCE_DATE_EPOCH environment variable for reproducible builds.
    See https://reproducible-builds.org/specs/source-date-epoch/
    """
    source_date_epoch = os.environ.get('SOURCE_DATE_EPOCH')
    if source_date_epoch is None:
        return datetime.now(timezone.utc)

    return datetime.fromtimestamp(int(source_date_epoch), timezone.utc)


def get_build_date():
    """
    Returns the displayable date string.

    Support SOURCE_DATE_EPOCH environment variable for reproducible builds.
    See https://reproducible-builds.org/specs/source-date-epoch/
    """
    return get_build_datetime().strftime('%Y-%m-%d')


def reduce_list(data_set):
    """ Reduce duplicate items in a list and preserve order """
    seen = set()
    return [item for item in data_set if
            item not in seen and not seen.add(item)]


def copy_file(source_path, output_path):
    """
    Copy source_path to output_path, making sure any parent directories exist.

    The output_path may be a directory.
    """
    output_dir = os.path.dirname(output_path)
    os.makedirs(output_dir, exist_ok=True)
    if os.path.isdir(output_path):
        output_path = os.path.join(output_path, os.path.basename(source_path))
    shutil.copyfile(source_path, output_path)


def write_file(content, output_path):
    """
    Write content to output_path, making sure any parent directories exist.
    """
    output_dir = os.path.dirname(output_path)
    os.makedirs(output_dir, exist_ok=True)
    with open(output_path, 'wb') as f:
        f.write(content)


def clean_directory(directory):
    """
    Remove the content of a directory recursively but not the directory itself.
    """
    if not os.path.exists(directory):
        return

    for entry in os.listdir(directory):

        # Don't remove hidden files from the directory. We never copy files
        # that are hidden, so we shouldn't delete them either.
        if entry.startswith('.'):
            continue

        path = os.path.join(directory, entry)
        if os.path.isdir(path):
            shutil.rmtree(path, True)
        else:
            os.unlink(path)


def get_html_path(path):
    """
    Map a source file path to an output html path.

    Paths like 'index.md' will be converted to 'index.html'
    Paths like 'about.md' will be converted to 'about/index.html'
    Paths like 'api-guide/core.md' will be converted to 'api-guide/core/index.html'
    """
    path = os.path.splitext(path)[0]
    if os.path.basename(path) == 'index':
        return path + '.html'
    return "/".join((path, 'index.html'))


def get_url_path(path, use_directory_urls=True):
    """
    Map a source file path to an output html path.

    Paths like 'index.md' will be converted to '/'
    Paths like 'about.md' will be converted to '/about/'
    Paths like 'api-guide/core.md' will be converted to '/api-guide/core/'

    If `use_directory_urls` is `False`, returned URLs will include the a trailing
    `index.html` rather than just returning the directory path.
    """
    path = get_html_path(path)
    url = '/' + path.replace(os.path.sep, '/')
    if use_directory_urls:
        return url[:-len('index.html')]
    return url


def is_markdown_file(path):
    """
    Return True if the given file path is a Markdown file.

    https://superuser.com/questions/249436/file-extension-for-markdown-files
    """
    return any(fnmatch.fnmatch(path.lower(), f'*{x}') for x in markdown_extensions)


def is_html_file(path):
    """
    Return True if the given file path is an HTML file.
    """
    ext = os.path.splitext(path)[1].lower()
    return ext in [
        '.html',
        '.htm',
    ]


def is_template_file(path):
    """
    Return True if the given file path is an HTML file.
    """
    ext = os.path.splitext(path)[1].lower()
    return ext in [
        '.html',
        '.htm',
        '.xml',
    ]


_ERROR_TEMPLATE_RE = re.compile(r'^\d{3}\.html?$')


def is_error_template(path):
    """
    Return True if the given file path is an HTTP error template.
    """
    return bool(_ERROR_TEMPLATE_RE.match(path))


@functools.lru_cache(maxsize=None)
def _norm_parts(path):
    if not path.startswith('/'):
        path = '/' + path
    path = posixpath.normpath(path)[1:]
    return path.split('/') if path else []


def get_relative_url(url, other):
    """
    Return given url relative to other.

    Both are operated as slash-separated paths, similarly to the 'path' part of a URL.
    The last component of `other` is skipped if it contains a dot (considered a file).
    Actual URLs (with schemas etc.) aren't supported. The leading slash is ignored.
    Paths are normalized ('..' works as parent directory), but going higher than the
    root has no effect ('foo/../../bar' ends up just as 'bar').
    """
    # Remove filename from other url if it has one.
    dirname, _, basename = other.rpartition('/')
    if '.' in basename:
        other = dirname

    other_parts = _norm_parts(other)
    dest_parts = _norm_parts(url)
    common = 0
    for a, b in zip(other_parts, dest_parts):
        if a != b:
            break
        common += 1

    rel_parts = ['..'] * (len(other_parts) - common) + dest_parts[common:]
    relurl = '/'.join(rel_parts) or '.'
    return relurl + '/' if url.endswith('/') else relurl


def normalize_url(path, page=None, base=''):
    """ Return a URL relative to the given page or using the base. """
    path, is_abs = _get_norm_url(path)
    if is_abs:
        return path
    if page is not None:
        return get_relative_url(path, page.url)
    return posixpath.join(base, path)


@functools.lru_cache(maxsize=None)
def _get_norm_url(path):
    path = path_to_url(path or '.')
    # Allow links to be fully qualified URLs
    parsed = urlsplit(path)
    if parsed.scheme or parsed.netloc or path.startswith(('/', '#')):
        return path, True
    return path, False


def create_media_urls(path_list, page=None, base=''):
    """
    Return a list of URLs relative to the given page or using the base.
    """
    return [normalize_url(path, page, base) for path in path_list]


def path_to_url(path):
    """Convert a system path to a URL."""

    return '/'.join(path.split('\\'))


def get_theme_dir(name):
    """ Return the directory of an installed theme by name. """

    theme = get_themes()[name]
    return os.path.dirname(os.path.abspath(theme.load().__file__))


def get_themes():
    """ Return a dict of all installed themes as {name: EntryPoint}. """

    themes = {}
    eps = set(importlib_metadata.entry_points(group='mkdocs.themes'))
    builtins = {ep.name for ep in eps if ep.dist.name == 'mkdocs'}

    for theme in eps:

        if theme.name in builtins and theme.dist.name != 'mkdocs':
            raise exceptions.ConfigurationError(
                f"The theme '{theme.name}' is a builtin theme but the package '{theme.dist.name}' "
                "attempts to provide a theme with the same name."
            )
        elif theme.name in themes:
            log.warning(
                f"A theme named '{theme.name}' is provided by the Python packages '{theme.dist.name}' "
                f"and '{themes[theme.name].dist.name}'. The one in '{theme.dist.name}' will be used."
            )

        themes[theme.name] = theme

    return themes


def get_theme_names():
    """Return a list of all installed themes by name."""

    return get_themes().keys()


def dirname_to_title(dirname):
    """ Return a page tile obtained from a directory name. """
    title = dirname
    title = title.replace('-', ' ').replace('_', ' ')
    # Capitalize if the dirname was all lowercase, otherwise leave it as-is.
    if title.lower() == title:
        title = title.capitalize()

    return title


def get_markdown_title(markdown_src):
    """
    Get the title of a Markdown document. The title in this case is considered
    to be a H1 that occurs before any other content in the document.
    The procedure is then to iterate through the lines, stopping at the first
    non-whitespace content. If it is a title, return that, otherwise return
    None.
    """

    lines = markdown_src.replace('\r\n', '\n').replace('\r', '\n').split('\n')
    while lines:
        line = lines.pop(0).strip()
        if not line.strip():
            continue
        if not line.startswith('# '):
            return
        return line.lstrip('# ')


def find_or_create_node(branch, key):
    """
    Given a list, look for dictionary with a key matching key and return it's
    value. If it doesn't exist, create it with the value of an empty list and
    return that.
    """

    for node in branch:
        if not isinstance(node, dict):
            continue

        if key in node:
            return node[key]

    new_branch = []
    node = {key: new_branch}
    branch.append(node)
    return new_branch


def nest_paths(paths):
    """
    Given a list of paths, convert them into a nested structure that will match
    the pages config.
    """
    nested = []

    for path in paths:

        if os.path.sep not in path:
            nested.append(path)
            continue

        directory, _ = os.path.split(path)
        parts = directory.split(os.path.sep)

        branch = nested
        for part in parts:
            part = dirname_to_title(part)
            branch = find_or_create_node(branch, part)

        branch.append(path)

    return nested


class CountHandler(logging.NullHandler):
    """ Counts all logged messages >= level. """

    def __init__(self, **kwargs):
        self.counts = defaultdict(int)
        super().__init__(**kwargs)

    def handle(self, record):
        rv = self.filter(record)
        if rv:
            # Use levelno for keys so they can be sorted later
            self.counts[record.levelno] += 1
        return rv

    def get_counts(self):
        return [(logging.getLevelName(k), v) for k, v in sorted(self.counts.items(), reverse=True)]


# For backward compatibility as some plugins import it.
# It is no longer necessary as all messages on the
# `mkdocs` logger get counted automatically.
warning_filter = logging.Filter()



try:
    from mkdocs.commands.babel import (
        compile_catalog,
        extract_messages,
        init_catalog,
        update_catalog
    )
    babel_cmdclass = {
        'compile_catalog': compile_catalog,
        'extract_messages': extract_messages,
        'init_catalog': init_catalog,
        'update_catalog': update_catalog,
    }
except ImportError:
    babel_cmdclass = {}



from click import ClickException, echo


class MkDocsException(ClickException):
    """Base exceptions for all MkDocs Exceptions"""


class Abort(MkDocsException):
    """Abort the build"""
    def show(self, **kwargs):
        echo(self.format_message())


class ConfigurationError(MkDocsException):
    """Error in configuration"""


class BuildError(MkDocsException):
    """Error during the build process"""


class PluginError(BuildError):
    """Error in a plugin"""



import unittest
from distutils.dist import Distribution
from distutils.errors import DistutilsOptionError
from os import path

from mkdocs.commands import babel

BASE_DIR = path.normpath(path.join(path.abspath(path.dirname(__file__)), '../../'))


class ThemeMixinTests(unittest.TestCase):

    def test_dict_entry_point(self):
        inst = babel.ThemeMixin()
        inst.distribution = Distribution()
        inst.distribution.entry_points = {
            'mkdocs.themes': [
                'mkdocs = mkdocs.themes.mkdocs'
            ]
        }
        inst.theme = 'mkdocs'
        self.assertEqual(inst.get_theme_dir(), path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs'))

    def test_ini_entry_point(self):
        inst = babel.ThemeMixin()
        inst.distribution = Distribution()
        inst.distribution.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        inst.theme = 'mkdocs'
        self.assertEqual(inst.get_theme_dir(), path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs'))

    def test_one_entry_point_as_default(self):
        inst = babel.ThemeMixin()
        inst.distribution = Distribution()
        inst.distribution.entry_points = {
            'mkdocs.themes': [
                'mkdocs = mkdocs.themes.mkdocs'
            ]
        }
        inst.theme = None
        self.assertEqual(inst.get_theme_dir(), path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs'))

    def test_multiple_entry_points(self):
        inst = babel.ThemeMixin()
        inst.distribution = Distribution()
        inst.distribution.entry_points = {
            'mkdocs.themes': [
                'mkdocs = mkdocs.themes.mkdocs',
                'readthedocs = mkdocs.themes.readthedocs',
            ]
        }
        inst.theme = 'readthedocs'
        self.assertEqual(inst.get_theme_dir(), path.join(BASE_DIR, 'mkdocs', 'themes', 'readthedocs'))

    def test_multiple_entry_points_no_default(self):
        inst = babel.ThemeMixin()
        inst.distribution = Distribution()
        inst.distribution.entry_points = {
            'mkdocs.themes': [
                'mkdocs = mkdocs.themes.mkdocs',
                'readthedocs = mkdocs.themes.readthedocs',
            ]
        }
        inst.theme = None
        self.assertRaises(DistutilsOptionError, inst.get_theme_dir)

    def test_no_entry_points(self):
        inst = babel.ThemeMixin()
        inst.distribution = Distribution()
        inst.distribution.entry_points = {}
        inst.theme = 'mkdocs'
        self.assertRaises(DistutilsOptionError, inst.get_theme_dir)

    def test_undefined_entry_point(self):
        inst = babel.ThemeMixin()
        inst.distribution = Distribution()
        inst.distribution.entry_points = {
            'mkdocs.themes': [
                'mkdocs = mkdocs.themes.mkdocs'
            ]
        }
        inst.theme = 'undefined'
        self.assertRaises(DistutilsOptionError, inst.get_theme_dir)


class CommandTests(unittest.TestCase):

    def test_compile_catalog(self):
        dist = Distribution()
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.compile_catalog(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.finalize_options()
        self.assertEqual(cmd.directory, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/locales'))

    def test_compile_catalog_default_theme(self):
        dist = Distribution()
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.compile_catalog(dist)
        cmd.initialize_options()
        self.assertIsNone(cmd.theme)
        cmd.finalize_options()
        self.assertEqual(cmd.theme, 'mkdocs')
        self.assertEqual(cmd.directory, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/locales'))

    def test_compile_catalog_ignore_theme(self):
        dist = Distribution()
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.compile_catalog(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.directory = 'foo/bar'
        cmd.finalize_options()
        self.assertEqual(cmd.directory, 'foo/bar')

    def test_extract_messages(self):
        dist = Distribution(dict(name='foo', version='1.2'))
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.extract_messages(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.finalize_options()
        self.assertEqual(cmd.input_paths, [path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs')])
        self.assertEqual(cmd.output_file, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/messages.pot'))
        self.assertEqual(cmd.mapping_file, babel.DEFAULT_MAPPING_FILE)
        self.assertEqual(cmd.project, 'foo')
        self.assertEqual(cmd.version, '1.2')

    def test_extract_messages_default_theme(self):
        dist = Distribution(dict(name='foo', version='1.2'))
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.extract_messages(dist)
        cmd.initialize_options()
        self.assertIsNone(cmd.theme)
        cmd.finalize_options()
        self.assertEqual(cmd.theme, 'mkdocs')
        self.assertEqual(cmd.input_paths, [path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs')])
        self.assertEqual(cmd.output_file, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/messages.pot'))

    def test_extract_messages_ingore_theme(self):
        dist = Distribution(dict(name='foo', version='1.2'))
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.extract_messages(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.input_paths = 'mkdocs/tests'
        cmd.output_file = 'foo/bar/messages.pot'
        cmd.finalize_options()
        self.assertEqual(cmd.input_paths, ['mkdocs/tests'])
        self.assertEqual(cmd.output_file, 'foo/bar/messages.pot')

    def test_extract_messages_ingore_theme_for_input(self):
        dist = Distribution(dict(name='foo', version='1.2'))
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.extract_messages(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.input_paths = 'mkdocs/tests'
        cmd.finalize_options()
        self.assertEqual(cmd.input_paths, ['mkdocs/tests'])
        self.assertEqual(cmd.output_file, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/messages.pot'))

    def test_extract_messages_ingore_theme_for_output(self):
        dist = Distribution(dict(name='foo', version='1.2'))
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.extract_messages(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.output_file = 'foo/bar/messages.pot'
        cmd.finalize_options()
        self.assertEqual(cmd.input_paths, [path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs')])
        self.assertEqual(cmd.output_file, 'foo/bar/messages.pot')

    def test_init_catalog(self):
        dist = Distribution()
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.init_catalog(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.locale = 'en'
        cmd.finalize_options()
        self.assertEqual(cmd.input_file, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/messages.pot'))
        self.assertEqual(cmd.output_dir, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/locales'))

    def test_init_catalog_default_theme(self):
        dist = Distribution()
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.init_catalog(dist)
        cmd.initialize_options()
        cmd.locale = 'en'
        self.assertIsNone(cmd.theme)
        cmd.finalize_options()
        self.assertEqual(cmd.theme, 'mkdocs')
        self.assertEqual(cmd.input_file, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/messages.pot'))
        self.assertEqual(cmd.output_dir, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/locales'))

    def test_init_catalog_ignore_theme(self):
        dist = Distribution()
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.init_catalog(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.locale = 'en'
        cmd.input_file = 'mkdocs/themes/mkdocs/messages.pot'
        cmd.output_dir = 'foo/bar'
        cmd.finalize_options()
        self.assertEqual(cmd.input_file, 'mkdocs/themes/mkdocs/messages.pot')
        self.assertEqual(cmd.output_dir, 'foo/bar')

    def test_init_catalog_ignore_theme_for_input(self):
        dist = Distribution()
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.init_catalog(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.locale = 'en'
        cmd.input_file = 'mkdocs/themes/mkdocs/messages.pot'
        cmd.finalize_options()
        self.assertEqual(cmd.input_file, 'mkdocs/themes/mkdocs/messages.pot')
        self.assertEqual(cmd.output_dir, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/locales'))

    def test_init_catalog_ignore_theme_for_output(self):
        dist = Distribution()
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.init_catalog(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.locale = 'en'
        cmd.output_dir = 'foo/bar'
        cmd.finalize_options()
        self.assertEqual(cmd.input_file, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/messages.pot'))
        self.assertEqual(cmd.output_dir, 'foo/bar')

    def test_update_catalog(self):
        dist = Distribution()
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.update_catalog(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.finalize_options()
        self.assertEqual(cmd.input_file, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/messages.pot'))
        self.assertEqual(cmd.output_dir, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/locales'))

    def test_update_catalog_default_theme(self):
        dist = Distribution()
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.update_catalog(dist)
        cmd.initialize_options()
        cmd.locale = 'en'
        self.assertIsNone(cmd.theme)
        cmd.finalize_options()
        self.assertEqual(cmd.theme, 'mkdocs')
        self.assertEqual(cmd.input_file, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/messages.pot'))
        self.assertEqual(cmd.output_dir, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/locales'))

    def test_update_catalog_ignore_theme(self):
        dist = Distribution()
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.update_catalog(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.locale = 'en'
        cmd.input_file = 'mkdocs/themes/readthedocs/messages.pot'
        cmd.output_dir = 'foo/bar'
        cmd.finalize_options()
        self.assertEqual(cmd.input_file, 'mkdocs/themes/readthedocs/messages.pot')
        self.assertEqual(cmd.output_dir, 'foo/bar')

    def test_update_catalog_ignore_theme_for_input(self):
        dist = Distribution()
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.update_catalog(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.locale = 'en'
        cmd.input_file = 'mkdocs/themes/mkdocs/messages.pot'
        cmd.finalize_options()
        self.assertEqual(cmd.input_file, 'mkdocs/themes/mkdocs/messages.pot')
        self.assertEqual(cmd.output_dir, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/locales'))

    def test_update_catalog_ignore_theme_for_output(self):
        dist = Distribution()
        dist.entry_points = '''
            [mkdocs.themes]
            mkdocs = mkdocs.themes.mkdocs
        '''
        cmd = babel.update_catalog(dist)
        cmd.initialize_options()
        cmd.theme = 'mkdocs'
        cmd.locale = 'en'
        cmd.output_dir = 'foo/bar'
        cmd.finalize_options()
        self.assertEqual(cmd.input_file, path.join(BASE_DIR, 'mkdocs', 'themes', 'mkdocs/messages.pot'))
        self.assertEqual(cmd.output_dir, 'foo/bar')
